"""
From https://github.com/pyg-team/pytorch_geometric/blob/master/examples/tgn.py
"""


# This code achieves a performance of around 96.60%. However, it is not
# directly comparable to the results reported by the TGN paper since a
# slightly different evaluation setup is used here.
# In particular, predictions in the same batch are made in parallel, i.e.
# predictions for interactions later in the batch have no access to any
# information whatsoever about previous interactions in the same batch.
# On the contrary, when sampling node neighborhoods for interactions later in
# the batch, the TGN paper code has access to previous interactions in the
# batch.
# While both approaches are correct, together with the authors of the paper we
# decided to present this version here as it is more realsitic and a better
# test bed for future methods.

import os.path as osp
import copy

import torch
from sklearn.metrics import average_precision_score, roc_auc_score
from torch.nn import Linear

from torch_geometric.datasets import JODIEDataset
from torch_geometric.loader import TemporalDataLoader
from torch_geometric.nn import TGNMemory, TransformerConv
from torch_geometric.nn.models.tgn import (
    IdentityMessage,
    LastAggregator,
    LastNeighborLoader,
)

import wandb

from utils import get_config_tgn, setup_seed, train_val1_val2_test_split, save_tgn
from bikeshare import get_cbs_data
from get_synth_data import get_synthetic_data

print("START")
# init wandb logger
wandb.init(
    project="generative-tgn",
    config=dict(
        name=get_config_tgn("name"),
        num_neighbors=get_config_tgn("num_neighbors"),
        batch_size=get_config_tgn("batch_size"),
        memory_dim=get_config_tgn("memory_dim"),
        time_dim=get_config_tgn("time_dim"),
        embedding_dim=get_config_tgn("embedding_dim"),
        data_name=get_config_tgn("data_name"),
        data_path=get_config_tgn("data_path"),
        lr=get_config_tgn("lr"),
        num_epochs=get_config_tgn("num_epochs"),
        seed=get_config_tgn("seed"),
        mode=get_config_tgn("mode"),
    ),
)

hyperparams = wandb.config

# setup seed
setup_seed(hyperparams["seed"])


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

name = hyperparams["data_name"]
print(f"Using {name} dataset. ")

data = None
if hyperparams["data_name"] in ["wikipedia", "reddit"]:
    path = hyperparams["data_path"]
    dataset = JODIEDataset(path, name=hyperparams["data_name"])
    data = dataset[0]

else:
    print("for sure bike data!!")
    data = get_cbs_data()

# For small datasets, we can put the whole dataset on GPU and thus avoid
# expensive memory transfer costs for mini-batches:
data = data.to(device)
data.msg = data.msg[:, :12]
print(f"DEBUG: {data.msg.shape[1]}")

# Ensure to only sample actual destination nodes as negatives.
min_dst_idx, max_dst_idx = int(data.dst.min()), int(data.dst.max())
train_data, val_1_data, val_2_data, test_data = train_val1_val2_test_split(data)

if hyperparams["mode"] == "eval_synth":
    val_1_data = get_synthetic_data()
    # val_1_data = torch.load("./small/reddit_bs10_nb10000")
    print("In synthetic eval mode!!!!")

batch_size = hyperparams["batch_size"]
train_loader = TemporalDataLoader(train_data, batch_size=batch_size)
val_1_loader = TemporalDataLoader(val_1_data, batch_size=batch_size)
val_2_loader = TemporalDataLoader(val_2_data, batch_size=batch_size)
test_loader = TemporalDataLoader(test_data, batch_size=batch_size)

neighbor_loader = LastNeighborLoader(
    data.num_nodes, size=hyperparams["num_neighbors"], device=device
)


class GraphAttentionEmbedding(torch.nn.Module):
    def __init__(self, in_channels, out_channels, msg_dim, time_enc):
        super().__init__()
        self.time_enc = time_enc
        edge_dim = msg_dim + time_enc.out_channels
        self.conv = TransformerConv(
            in_channels, out_channels // 2, heads=2, dropout=0.1, edge_dim=edge_dim
        )

    def forward(self, x, last_update, edge_index, t, msg):
        rel_t = last_update[edge_index[0]] - t
        rel_t_enc = self.time_enc(rel_t.to(x.dtype))
        edge_attr = torch.cat([rel_t_enc, msg], dim=-1)
        return self.conv(x, edge_index, edge_attr)


class LinkPredictor(torch.nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.lin_src = Linear(in_channels, in_channels)
        self.lin_dst = Linear(in_channels, in_channels)
        self.lin_final = Linear(in_channels, 1)

    def forward(self, z_src, z_dst):
        h = self.lin_src(z_src) + self.lin_dst(z_dst)
        h = h.relu()
        return self.lin_final(h)


memory_dim = hyperparams["memory_dim"]
time_dim = hyperparams["time_dim"]
embedding_dim = hyperparams["embedding_dim"]

memory = TGNMemory(
    data.num_nodes,
    data.msg.size(-1),
    memory_dim,
    time_dim,
    message_module=IdentityMessage(data.msg.size(-1), memory_dim, time_dim),
    aggregator_module=LastAggregator(),
).to(device)

gnn = GraphAttentionEmbedding(
    in_channels=memory_dim,
    out_channels=embedding_dim,
    msg_dim=data.msg.size(-1),
    time_enc=memory.time_enc,
).to(device)

link_pred = LinkPredictor(in_channels=embedding_dim).to(device)

optimizer = torch.optim.Adam(
    set(memory.parameters()) | set(gnn.parameters()) | set(link_pred.parameters()),
    lr=hyperparams["lr"],
)
criterion = torch.nn.BCEWithLogitsLoss()

# Helper vector to map global node indices to local ones.
assoc = torch.empty(data.num_nodes, dtype=torch.long, device=device)


def train():
    memory.train()
    gnn.train()
    link_pred.train()

    memory.reset_state()  # Start with a fresh memory.
    neighbor_loader.reset_state()  # Start with an empty graph.

    total_loss = 0
    for batch in train_loader:
        batch = batch.to(device)
        optimizer.zero_grad()

        src, pos_dst, t, msg = batch.src, batch.dst, batch.t, batch.msg
        # print("FOR TRAIN!!")
        # print(src.shape)
        # print(pos_dst.shape)
        # print(t.shape)
        # print(msg.shape)

        # Sample negative destination nodes.
        neg_dst = torch.randint(
            min_dst_idx,
            max_dst_idx + 1,
            (src.size(0),),
            dtype=torch.long,
            device=device,
        )

        n_id = torch.cat([src, pos_dst, neg_dst]).unique()
        n_id, edge_index, e_id = neighbor_loader(n_id)
        assoc[n_id] = torch.arange(n_id.size(0), device=device)

        # Get updated memory of all nodes involved in the computation.
        z, last_update = memory(n_id)
        z = gnn(
            z,
            last_update,
            edge_index,
            data.t[e_id].to(device),
            data.msg[e_id].to(device),
        )

        pos_out = link_pred(z[assoc[src]], z[assoc[pos_dst]])
        neg_out = link_pred(z[assoc[src]], z[assoc[neg_dst]])

        loss = criterion(pos_out, torch.ones_like(pos_out))
        loss += criterion(neg_out, torch.zeros_like(neg_out))

        # Update memory and neighbor loader with ground-truth state.
        memory.update_state(src, pos_dst, t, msg)
        neighbor_loader.insert(src, pos_dst)

        loss.backward()
        optimizer.step()
        memory.detach()
        total_loss += float(loss) * batch.num_events

    return total_loss / train_data.num_events


@torch.no_grad()
def test(loader):
    memory.eval()
    gnn.eval()
    link_pred.eval()

    torch.manual_seed(
        hyperparams["seed"]
    )  # Ensure deterministic sampling across epochs.

    aps, aucs = [], []
    for batch in loader:
        batch = batch.to(device)
        src, pos_dst, t, msg = batch.src, batch.dst, batch.t, batch.msg

        # print("FOR TEST!!!!")
        # print(src.shape)
        # print(pos_dst.shape)
        # print(t.shape)
        # print(msg.shape)

        neg_dst = torch.randint(
            min_dst_idx,
            max_dst_idx + 1,
            (src.size(0),),
            dtype=torch.long,
            device=device,
        )

        n_id = torch.cat([src, pos_dst, neg_dst]).unique()
        n_id, edge_index, e_id = neighbor_loader(n_id)
        assoc[n_id] = torch.arange(n_id.size(0), device=device)

        z, last_update = memory(n_id)
        z = gnn(
            z,
            last_update,
            edge_index,
            data.t[e_id].to(device),
            data.msg[e_id].to(device),
        )

        pos_out = link_pred(z[assoc[src]], z[assoc[pos_dst]])
        neg_out = link_pred(z[assoc[src]], z[assoc[neg_dst]])

        y_pred = torch.cat([pos_out, neg_out], dim=0).sigmoid().cpu()
        y_true = torch.cat(
            [torch.ones(pos_out.size(0)), torch.zeros(neg_out.size(0))], dim=0
        )

        aps.append(average_precision_score(y_true, y_pred))
        aucs.append(roc_auc_score(y_true, y_pred))

        memory.update_state(src, pos_dst, t, msg)
        neighbor_loader.insert(src, pos_dst)

    return float(torch.tensor(aps).mean()), float(torch.tensor(aucs).mean())


best_val_ap = 0
# best_memory = None
# best_gnn = None
# best_link_pred = None
for epoch in range(1, hyperparams["num_epochs"] + 1):
    loss = train()
    print(f"Epoch: {epoch:02d}, Loss: {loss:.4f}")
    val_ap, val_auc = test(val_1_loader)
    # test_ap, test_auc = test(test_loader)
    print(f"Val AP: {val_ap:.4f}, Val AUC: {val_auc:.4f}")
    # print(f"Test AP: {test_ap:.4f}, Test AUC: {test_auc:.4f}")

    if val_ap > best_val_ap:
        best_val_ap = val_ap
        best_memory = copy.deepcopy(memory)
        best_gnn = copy.deepcopy(gnn)
        best_link_pred = copy.deepcopy(link_pred)

    wandb.log(
        {
            "train_loss": loss,
            "val_ap": val_ap,
            "val_auc": val_auc,
            "best_val_ap": best_val_ap,
        }
    )


print("Running best model on second held out val set")
memory = best_memory
gnn = best_gnn
link_pred = best_link_pred
test_ap, test_auc = test(val_2_loader)

wandb.run.summary["best_model_val_2_ap"] = test_ap
wandb.run.summary["best_model_val_2_auc"] = test_auc
print("DONE!!")
